"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.unpack = exports.createHashMapping = exports.unpackObj = exports.unpackArray = exports.pack = exports.listKeys = exports.SDJwt = void 0;
const crypto_1 = require("./crypto");
const decoy_1 = require("./decoy");
const disclosure_1 = require("./disclosure");
const error_1 = require("./error");
const jwt_1 = require("./jwt");
const kbjwt_1 = require("./kbjwt");
const type_1 = require("./type");
class SDJwt {
    constructor(data) {
        this.jwt = data === null || data === void 0 ? void 0 : data.jwt;
        this.disclosures = data === null || data === void 0 ? void 0 : data.disclosures;
        this.kbJwt = data === null || data === void 0 ? void 0 : data.kbJwt;
    }
    static decodeSDJwt(sdjwt) {
        const [encodedJwt, ...encodedDisclosures] = sdjwt.split(type_1.SD_SEPARATOR);
        const jwt = jwt_1.Jwt.fromEncode(encodedJwt);
        if (encodedDisclosures.length === 0) {
            return {
                jwt,
                disclosures: [],
            };
        }
        const encodedKeyBindingJwt = encodedDisclosures.pop();
        const kbJwt = encodedKeyBindingJwt
            ? kbjwt_1.KBJwt.fromKBEncode(encodedKeyBindingJwt)
            : undefined;
        const disclosures = encodedDisclosures.map(disclosure_1.Disclosure.fromEncode);
        return {
            jwt,
            disclosures,
            kbJwt,
        };
    }
    static fromEncode(encodedSdJwt) {
        const { jwt, disclosures, kbJwt } = SDJwt.decodeSDJwt(encodedSdJwt);
        return new SDJwt({
            jwt,
            disclosures,
            kbJwt,
        });
    }
    present(keys) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            if (!((_a = this.jwt) === null || _a === void 0 ? void 0 : _a.payload) || !this.disclosures) {
                throw new error_1.SDJWTException('Invalid sd-jwt: jwt or disclosures is missing');
            }
            const hashmap = yield (0, exports.createHashMapping)(this.disclosures);
            const { disclosureKeymap } = yield (0, exports.unpack)((_b = this.jwt) === null || _b === void 0 ? void 0 : _b.payload, this.disclosures);
            const presentableKeys = Object.keys(disclosureKeymap);
            const missingKeys = keys.filter((k) => !presentableKeys.includes(k));
            if (missingKeys.length > 0) {
                throw new error_1.SDJWTException(`Invalid sd-jwt: invalid present keys: ${missingKeys.join(', ')}`);
            }
            const disclosures = keys.map((k) => hashmap[disclosureKeymap[k]]);
            const presentSDJwt = new SDJwt({
                jwt: this.jwt,
                disclosures,
                kbJwt: this.kbJwt,
            });
            return presentSDJwt.encodeSDJwt();
        });
    }
    encodeSDJwt() {
        const data = [];
        if (!this.jwt) {
            throw new error_1.SDJWTException('Invalid sd-jwt: jwt is missing');
        }
        const encodedJwt = this.jwt.encodeJwt();
        data.push(encodedJwt);
        if (this.disclosures && this.disclosures.length > 0) {
            const encodeddisclosures = this.disclosures
                .map((dc) => dc.encode())
                .join(type_1.SD_SEPARATOR);
            data.push(encodeddisclosures);
        }
        data.push(this.kbJwt ? this.kbJwt.encodeJwt() : '');
        return data.join(type_1.SD_SEPARATOR);
    }
    keys() {
        return __awaiter(this, void 0, void 0, function* () {
            return (0, exports.listKeys)(yield this.getClaims()).sort();
        });
    }
    presentableKeys() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            if (!((_a = this.jwt) === null || _a === void 0 ? void 0 : _a.payload) || !this.disclosures) {
                throw new error_1.SDJWTException('Invalid sd-jwt: jwt or disclosures is missing');
            }
            const { disclosureKeymap } = yield (0, exports.unpack)((_b = this.jwt) === null || _b === void 0 ? void 0 : _b.payload, this.disclosures);
            return Object.keys(disclosureKeymap).sort();
        });
    }
    getClaims() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            if (!((_a = this.jwt) === null || _a === void 0 ? void 0 : _a.payload) || !this.disclosures) {
                throw new error_1.SDJWTException('Invalid sd-jwt: jwt or disclosures is missing');
            }
            const { unpackedObj } = yield (0, exports.unpack)((_b = this.jwt) === null || _b === void 0 ? void 0 : _b.payload, this.disclosures);
            return unpackedObj;
        });
    }
}
exports.SDJwt = SDJwt;
const listKeys = (obj, prefix = '') => {
    const keys = [];
    for (let key in obj) {
        if (obj[key] === undefined)
            continue;
        const newKey = prefix ? `${prefix}.${key}` : key;
        keys.push(newKey);
        if (obj[key] && typeof obj[key] === 'object' && obj[key] !== null) {
            keys.push(...(0, exports.listKeys)(obj[key], newKey));
        }
    }
    return keys;
};
exports.listKeys = listKeys;
const pack = (claims, disclosureFrame, hasher = crypto_1.digest, saltGenerator = crypto_1.generateSalt) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b;
    if (!disclosureFrame) {
        return {
            packedClaims: claims,
            disclosures: [],
        };
    }
    const sd = (_a = disclosureFrame[type_1.SD_DIGEST]) !== null && _a !== void 0 ? _a : [];
    const decoyCount = (_b = disclosureFrame[type_1.SD_DECOY]) !== null && _b !== void 0 ? _b : 0;
    if (claims instanceof Array) {
        const packedClaims = [];
        const disclosures = [];
        const recursivePackedClaims = {};
        for (const key in disclosureFrame) {
            if (key !== type_1.SD_DIGEST) {
                const idx = parseInt(key);
                // @ts-ignore
                const packed = yield (0, exports.pack)(claims[idx], disclosureFrame[idx], hasher);
                recursivePackedClaims[idx] = packed.packedClaims;
                disclosures.push(...packed.disclosures);
            }
        }
        for (let i = 0; i < claims.length; i++) {
            const claim = recursivePackedClaims[i]
                ? recursivePackedClaims[i]
                : claims[i];
            // @ts-ignore
            if (sd.includes(i)) {
                const salt = saltGenerator(16);
                const disclosure = new disclosure_1.Disclosure([salt, claim]);
                const digest = yield disclosure.digest(hasher);
                packedClaims.push({ '...': digest });
                disclosures.push(disclosure);
            }
            else {
                packedClaims.push(claim);
            }
        }
        for (let j = 0; j < decoyCount; j++) {
            const decoyDigest = yield (0, decoy_1.createDecoy)(hasher, saltGenerator);
            packedClaims.push({ '...': decoyDigest });
        }
        return { packedClaims, disclosures };
    }
    const packedClaims = {};
    const disclosures = [];
    const recursivePackedClaims = {};
    for (const key in disclosureFrame) {
        if (key !== type_1.SD_DIGEST) {
            const packed = yield (0, exports.pack)(
            // @ts-ignore
            claims[key], disclosureFrame[key], hasher);
            recursivePackedClaims[key] = packed.packedClaims;
            disclosures.push(...packed.disclosures);
        }
    }
    const _sd = [];
    for (const key in claims) {
        const claim = recursivePackedClaims[key]
            ? recursivePackedClaims[key]
            : claims[key];
        // @ts-ignore
        if (sd.includes(key)) {
            const salt = saltGenerator(16);
            const disclosure = new disclosure_1.Disclosure([salt, key, claim]);
            const digest = yield disclosure.digest(hasher);
            _sd.push(digest);
            disclosures.push(disclosure);
        }
        else {
            packedClaims[key] = claim;
        }
    }
    for (let j = 0; j < decoyCount; j++) {
        const decoyDigest = yield (0, decoy_1.createDecoy)(hasher, saltGenerator);
        _sd.push(decoyDigest);
    }
    if (_sd.length > 0) {
        packedClaims[type_1.SD_DIGEST] = _sd.sort();
    }
    return { packedClaims, disclosures };
});
exports.pack = pack;
const unpackArray = (arr, map, prefix = '') => {
    const keys = {};
    const unpackedArray = [];
    arr.forEach((item, idx) => {
        if (item instanceof Object) {
            if (item[type_1.SD_LIST_KEY]) {
                const hash = item[type_1.SD_LIST_KEY];
                const disclosed = map[hash];
                if (disclosed) {
                    const presentKey = prefix ? `${prefix}.${idx}` : `${idx}`;
                    keys[presentKey] = hash;
                    const { unpackedObj, disclosureKeymap: disclosureKeys } = (0, exports.unpackObj)(disclosed.value, map, presentKey);
                    unpackedArray.push(unpackedObj);
                    Object.assign(keys, disclosureKeys);
                }
            }
            else {
                const newKey = prefix ? `${prefix}.${idx}` : `${idx}`;
                const { unpackedObj, disclosureKeymap: disclosureKeys } = (0, exports.unpackObj)(item, map, newKey);
                unpackedArray.push(unpackedObj);
                Object.assign(keys, disclosureKeys);
            }
        }
        else {
            unpackedArray.push(item);
        }
    });
    return { unpackedObj: unpackedArray, disclosureKeymap: keys };
};
exports.unpackArray = unpackArray;
const unpackObj = (obj, map, prefix = '') => {
    const keys = {};
    if (obj instanceof Object) {
        if (obj instanceof Array) {
            return (0, exports.unpackArray)(obj, map, prefix);
        }
        for (const key in obj) {
            if (key !== type_1.SD_DIGEST &&
                key !== type_1.SD_LIST_KEY &&
                obj[key] instanceof Object) {
                const newKey = prefix ? `${prefix}.${key}` : key;
                const { unpackedObj, disclosureKeymap: disclosureKeys } = (0, exports.unpackObj)(obj[key], map, newKey);
                obj[key] = unpackedObj;
                Object.assign(keys, disclosureKeys);
            }
        }
        const { _sd } = obj, payload = __rest(obj, ["_sd"]);
        const claims = {};
        if (_sd) {
            _sd.forEach((hash) => {
                const disclosed = map[hash];
                if (disclosed && disclosed.key) {
                    const presentKey = prefix
                        ? `${prefix}.${disclosed.key}`
                        : disclosed.key;
                    keys[presentKey] = hash;
                    const { unpackedObj, disclosureKeymap: disclosureKeys } = (0, exports.unpackObj)(disclosed.value, map, presentKey);
                    claims[disclosed.key] = unpackedObj;
                    Object.assign(keys, disclosureKeys);
                }
            });
        }
        const unpackedObj = Object.assign(payload, claims);
        return { unpackedObj, disclosureKeymap: keys };
    }
    return { unpackedObj: obj, disclosureKeymap: keys };
};
exports.unpackObj = unpackObj;
const createHashMapping = (disclosures, hasher = crypto_1.digest) => __awaiter(void 0, void 0, void 0, function* () {
    const map = {};
    for (let i = 0; i < disclosures.length; i++) {
        const disclosure = disclosures[i];
        const digest = yield disclosure.digest(hasher);
        map[digest] = disclosure;
    }
    return map;
});
exports.createHashMapping = createHashMapping;
const unpack = (sdjwtPayload, disclosures) => __awaiter(void 0, void 0, void 0, function* () {
    const { _sd_alg } = sdjwtPayload, payload = __rest(sdjwtPayload, ["_sd_alg"]);
    const hasher = (0, crypto_1.getHasher)(_sd_alg);
    const map = yield (0, exports.createHashMapping)(disclosures, hasher);
    return (0, exports.unpackObj)(payload, map);
});
exports.unpack = unpack;
